#!/usr/bin/python2
#
# A script that calculates the space occupied by files in every directory
# and can set the metadata filesize field with it, for the first N levels
#
# This script can directly update information in the database and not all
# operation leads to consistent results if you don't stop all DPM daemons
# before executing this script (see help screen for more details and examples)
#
# March 2015 - Fabrizio Furano - CERN IT/SDC - furano@cern.ch
# January 2019 - Petr Vokac - petr.vokac@cern.ch
#
#
# Usage:
# * get the help screen with all available options
#   dmlite-mysql-dirspaces [-h] [--help]
from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import sys, os
import optparse
import re
import stat
import datetime
import collections
import MySQLdb
import MySQLdb.cursors
import logging, logging.handlers

try: import simplejson as json
except ImportError: import json

__version__ = '1.1.0'
__author__ = 'Fabrizio Furano'

_log = logging.getLogger('DPMDIRSPACES')

DEFAULT_CNS_DB = 'cns_db'
DEFAULT_DPM_DB = 'dpm_db'
DEFAULT_DOME_CONFIG = '/etc/domehead.conf'
DEFAULT_UPDATELEVELS = 6



def guess_config_files():
    """ Guesses the location of DPM namespace configuration file """
    try:
        possible_nsconfigs = ['/opt/lcg/etc/NSCONFIG', '/usr/etc/NSCONFIG']
        if 'LCG_LOCATION' in os.environ:
            possible_nsconfigs.insert(0, os.environ['LCG_LOCATION'].rstrip('/') + '/etc/NSCONFIG')

        for f in possible_nsconfigs:
            if os.path.exists(f):
                return f

    except Exception as e:
        _log.warn("failed to guess DB config file location: %s", str(e))

    return None



def get_conn_data(nsconfig):
    """ Returns connection data from NSCONFIG"""
    retval = {}

    _log.debug("getting connection info from %s", nsconfig)
    try:
        nsconfig_line = open(nsconfig).readline().strip()
    except Exception as e:
        _log.error("Cannot open DPM config file %s: %s", nsconfig, str(e))
        sys.exit(-1)

    nsre = re.compile(r"(.*)/(.*)@([^/]*)(?:/(.*))?")
    m = nsre.match(nsconfig_line)
    if m == None:
        _log.error("Bad line in DPM config '%s', doesn't match re '%s'", nsconfig, nsre)
        sys.exit(-1)
    retval['user'] = m.group(1)
    retval['pass'] = m.group(2)
    retval['host'] = m.group(3)
    if m.group(4):
        retval['cns_db'] = m.group(4)
    else:
        retval['cns_db'] = DEFAULT_CNS_DB
    retval['dpm_db'] = DEFAULT_DPM_DB

    _log.debug("database connection: host=%s, user=%s, cns_db=%s, dpm_db=%s", retval['host'], retval['user'], retval['cns_db'], retval['dpm_db'])

    return retval



def get_connection(conn_data, db, cclass=MySQLdb.cursors.Cursor):
    return MySQLdb.connect(
            host=conn_data['host'], user=conn_data['user'],
            passwd=conn_data['pass'], db=conn_data[db],
            cursorclass=cclass)



def get_updatelivels(filename):
    """ Guesses the location of DPM namespace configuration file """
    ret = DEFAULT_UPDATELEVELS

    try:
        if os.path.exists(filename):
            with open(filename) as f:
                for line in f:
                    res = re.match('^head.dirspacereportdepth:\s*(\d+).*', line)
                    if res != None:
                        ret = int(res.group(1))
                        _log.debug("found updatelevel %i configuration in %s", ret, filename)
        else:
            _log.debug('DPM DOME head config %s not found', filename)

    except Exception as e:
        _log.warn("failed to guess updatelvels from %s: %s", filename, str(e))

    return ret



class CachedFullPath(object):
    """DPM file metadata stored in database have pointer just
    to parent directory and to build whole path it is necessary
    to recursively query all parent directories to the root "/".

    Althought these queries are pretty cheap they are done for
    each file and even with small latencies (especially with
    database on dedicated server) they can take quite a time.

    This class not only caches past queries, but also limit
    size of cached data not to exhaust whole memory while
    dumping big DPM database with a lot of files.
    """

    def __init__(self, conn, maxsize=1000000):
        self._cache = {}
        self._cache_rev = {}
        self._conn = conn
        self._maxsize = maxsize
        self._ntotal = 0
        self._ncached = 0
        self._nqueries = 0
        self._ncleanup = 0

    def __del__(self):
        if _log:
            _log.info("path lookup cache usage: total %i, cached %i, queries %i, cleanup %i", self._ntotal, self._ncached, self._nqueries, self._ncleanup)

    def get(self, fileid):
        """Recursively get full path for given fileid"""
        self._ntotal += 1
        if fileid == 0:
            return ""

        if fileid in self._cache:
            self._ncached += 1
            return self._cache[fileid]

        if len(self._cache) >= self._maxsize:
            _log.debug("fullpath cache too big (%i), dropping cached records...", len(self._cache))
            self._ncleanup += 1
            del(self._cache)
            del(self._cache_rev)
            self._cache = {}
            self._cache_rev = {}

        sql = "SELECT parent_fileid, name FROM Cns_file_metadata WHERE fileid=%s"
        cursor = self._conn.cursor()
        cursor.execute(sql, (fileid, ))
        res = cursor.fetchone()
        cursor.close()

        self._nqueries += 1

        if _log.getEffectiveLevel() < logging.DEBUG:
            _log.debug("query parent directory '%s': %s", sql, res)

        fullname = None
        if res:
            (parentid, name) = res

            if parentid == 0:
                name = ''

            if name.find('/') != -1: # this script doesn't support '/' characters in metadata name
                raise Exception("fileid {0} contains slash character in its name '{1}'".format(fileid, name))

            if parentid == 0:
                fullname = ''
            else:
                prefix = self.get(parentid)
                if prefix != None:
                    fullname = "{0}/{1}".format(prefix, name)
        else:
            _log.warn("db inconsistency: could not find fileid %i (most likely the entry is orphan).", fileid)

        self._cache[fileid] = fullname
        self._cache_rev[fullname] = fileid
        return fullname

    def get_multi(self, fileids):
        """Reduce impact of query latency by resolving paths for multiple
        fileids. Max number of queries still corresponds to the max path
        depth, but all fileids are resolved at the same time."""
        self._ntotal += len(fileids)

        if len(self._cache) + len(fileids) >= self._maxsize:
            _log.debug("fullpath cache too big (%i+%i), dropping cached records...", len(self._cache), len(fileids))
            self._ncleanup += 1
            del(self._cache)
            del(self._cache_rev)
            self._cache = {}
            self._cache_rev = {}

        tores = set()
        id2name = {}
        id2parent = {}
        for fileid in fileids:
            if fileid in self._cache:
                self._ncached += 1
            else:
                tores.add(str(fileid))

        if len(tores) > 0:
            cursor = self._conn.cursor()

            while len(tores) > 0:

                sql = "SELECT fileid, parent_fileid, name FROM Cns_file_metadata WHERE fileid IN ({0})".format(','.join(tores))
                cursor.execute(sql)

                self._nqueries += 1
                tores = set()

                for row in cursor:
                    (fileid, parentid, name) = row

                    if _log.getEffectiveLevel() < logging.DEBUG:
                        _log.debug("query parent directory '%s': %s", sql, row)

                    if parentid == 0:
                        name = ''

                    if name.find('/') != -1: # this script doesn't support '/' characters in metadata name
                        raise Exception("fileid {0} contains slash character in its name '{1}'".format(fileid, name))

                    id2name[fileid] = name
                    id2parent[fileid] = parentid

                    if parentid == 0:
                        pass
                    elif parentid in self._cache:
                        self._ncached += 1
                    elif parentid not in id2parent:
                        tores.add(str(parentid))

            cursor.close()

        ret = {}
        for fileid in fileids:
            if fileid in self._cache:
                ret[fileid] = self._cache[fileid]
            else:
                currid = fileid
                revname = []

                while True:
                    if currid in self._cache:
                        revname.append((currid, self._cache[currid]))
                        self._ncached += 1
                        break
                    elif currid in id2parent:
                        revname.append((currid, id2name[currid]))
                        currid = id2parent[currid]
                        if currid == 0:
                            break
                    else:
                        _log.warn("db inconsistency: could not find fileid %i parent %i %s (most likely the entry is orphan).", fileid, currid, str(revname))
                        revname = []
                        break

                if len(revname) > 0:
                    revname.reverse()
                    for i in range(len(revname)):
                        currid = revname[i][0]
                        if currid in self._cache: continue
                        fullname = '/'.join([ x[1] for x in revname[:i+1] ])
                        self._cache[currid] = fullname
                        self._cache_rev[fullname] = currid

                    ret[fileid] = self._cache[fileid]

        return ret


    def get_ids(self, fileid):
        path = self.get(fileid)
        if path == None: return None
        parts = path.split('/')
        return [ self._cache_rev['/'.join(parts[:i+1])] for i in range(len(parts)) ]


    def get_ids_multi(self, fileids):
        ret = {}

        paths = self.get_multi(fileids)
        for fileid, path in paths.items():
            if path == None: continue
            parts = path.split('/')
            ret[fileid] = [ self._cache_rev['/'.join(parts[:i+1])] for i in range(len(parts)) ]

        return ret



def get_path_size(conn, path):
    """Get size for given path"""
    size = -1
    cursor = conn.cursor()
    try:
        done = []
        parentid = 0
        for name in path.split('/'):
            if name == '': name = '/' # root directory
            cursor.execute('SELECT fileid, filesize from Cns_file_metadata WHERE parent_fileid = %s AND name = %s', (parentid, name))
            res = cursor.fetchone()
            if res == None:
                _log.warn("for path {0} could not find metadata for {1} (parentid: {2}, name: {3}), assuming 0 size".format(path, '/'.join(done), parentid, name))
                size = 0
                break
            parentid, size = res

    finally:
        cursor.close()

    return size



def get_quotatoken_data(conn):
    cursor = conn.cursor()

    ret = []
    try:
        cursor.execute("SELECT path, poolname, u_token, s_token, t_space, g_space, u_space FROM dpm_space_reserv")
        rows = cursor.fetchall()

        for path, poolname, u_token, s_token, t_space, g_space, u_space in rows:
            if path == None:
                _log.error("missing directory path for quotatoken %s", u_token)

        # sort quotatokens by path length with longest first
        # (this ordering is assumed by code calling this function)
        for row in sorted(filter(lambda x: x[0] != None, rows), cmp=lambda x, y: cmp(len(y[0]), len(x[0]))):
            path, poolname, u_token, s_token, t_space, g_space, u_space = row
            ret.append((path, poolname, u_token, s_token, t_space, g_space, u_space))

    finally:
        cursor.close()

    return ret



def get_accounted_dirs(conn, updatelevels):
    """find all directories up to updatelevels"""
    cursor = conn.cursor()
    try:

        cursor.execute('SELECT fileid FROM Cns_file_metadata WHERE parent_fileid = 0 AND (filemode & {0}) = {0}'.format(stat.S_IFDIR))
        fileid = cursor.fetchone()[0]

        adirids = [fileid]
        currfileids = [str(fileid)]

        for level in range(updatelevels):
            cursor.execute('SELECT fileid FROM Cns_file_metadata WHERE parent_fileid IN ({0}) AND (filemode & {1}) = {1}'.format(','.join(currfileids), stat.S_IFDIR))
            currfileids = []
            for row in cursor:
                fileid = row[0]
                adirids.append(fileid)
                currfileids.append(str(fileid))

        _log.info("found %i directories within %i levels", len(adirids), updatelevels)

    finally:
        cursor.close()

    return adirids



def fix_zero_dirs(conn_data, updatelevels, updatedb=False):
    """Set filesize to zero for directories below update level"""
    _log.debug("fix_zero_dirs(updatelevels=%i, updatedb=%s)", updatelevels, updatedb)

    conn = get_connection(conn_data, 'cns_db')
    cursor = conn.cursor()
    try:
        currfileids = get_accounted_dirs(conn, updatelevels)
        currfileids_list = ','.join([ str(x) for x in currfileids ])

        if updatedb:
            cursor.execute('UPDATE Cns_file_metadata SET filesize = 0 WHERE fileid NOT IN ({0}) AND (filemode & {1}) = {1} AND filesize != 0'.format(currfileids_list, stat.S_IFDIR))
            conn.commit()
        else:
            pathname = CachedFullPath(conn)
            cursor.execute('SELECT fileid, filesize FROM Cns_file_metadata WHERE fileid NOT IN ({0}) AND (filemode & {1}) = {1} AND filesize != 0'.format(currfileids_list, stat.S_IFDIR))
            for row in cursor:
                fileid, filesize = row
                _log.info("dry-run: update directory %s (below %i levels) with size %i to zero", pathname.get(fileid), updatelevels, filesize)

        _log.info("updated %i directories below %i levels to filesize zero", cursor.rowcount, updatelevels)

    finally:
        cursor.close()



def fix_spacetokens_by_path(conn_data, skip=[], updatedb=False):
    # NOTE: this function doesn't check if storage pool associated with replica
    # use right host:/fs defined in 'dpm_db'.'dpm_fs' table
    _log.debug("fix_spacetokens_by_path(skip=%s, updatedb=%s)", skip, updatedb)

    dry_run = '' if updatedb else 'dry-run: '

    conn_dpm = get_connection(conn_data, 'dpm_db')
    qt = get_quotatoken_data(conn_dpm)
    s2name = {}
    for path, poolname, u_token, s_token, t_space, g_space, u_space in qt:
        s2name[s_token] = u_token
    conn_dpm.close()

    conn = get_connection(conn_data, 'cns_db')
    conn_ss = get_connection(conn_data, 'cns_db', cclass=MySQLdb.cursors.SSCursor)
    conn_path = get_connection(conn_data, 'cns_db')
    pathname = CachedFullPath(conn_path)

    try:
        # join metadata table with replicas to account size of all file replicas
        cursor = conn_ss.cursor()
        ucursor = conn.cursor()

        sql = 'SELECT replica.rowid, replica.fileid, metadata.parent_fileid, metadata.name, replica.poolname, replica.setname, replica.xattr FROM Cns_file_metadata AS metadata JOIN Cns_file_replica AS replica ON metadata.fileid = replica.fileid'
        cursor.execute(sql)

        updated = 0
        cnt_rows = 0
        while True:
            # retreiving data in chunks dramatically increase performance
            # mostly because of latency associated with each DB query
            rows = cursor.fetchmany(1000)
            if len(rows) == 0: break

            if _log.getEffectiveLevel() < logging.DEBUG:
                _log.debug("fetched %i rows", len(rows))

            fileids = [ row[2] for row in rows ]
            pathnames = pathname.get_multi(fileids)

            for row in rows:
                cnt_rows += 1

                if cnt_rows % 1000000 == 0:
                    _log.debug("processed %i records, updated %i", cnt_rows, updated)

                if _log.getEffectiveLevel() < logging.DEBUG:
                    _log.debug("row %i: %s", cnt_rows, str(row))

                rowid, fileid, pfileid, name, poolname, setname, xattr = row
                path = pathnames.get(pfileid)
                if path == None:
                    _log.error("skipping fileid %i, unable to reconstruct its parent %i path", fileid, pfileid)
                    continue

                qt_selected = filter(lambda x: path.startswith(x[0]), qt)
                if len(qt_selected) == 0:
                    _log.warn("skipping %s/%s with fileid %i, because it doen't match any quotatoken path", path, name, fileid)
                    continue

                # first selected QT is the right one because "qt" array is sorted by path length
                spath, spool, u_token, s_token, t_space, g_space, u_space = qt_selected[0]
                # skip replicas with matching pool and spacetokens
                if poolname == spool and setname == s_token: continue
                # update only spacetokens not explicetely excluded
                if setname in skip and s_token in skip: continue

                # deal with JSON in xattr
                xattr_dict = {}
                if xattr != None and xattr not in [ None, '', '{}' ]:
                    try:
                        xattr_dict = json.loads(xattr)
                    except Exception as e:
                        _log.error("unable to parse replica xattr for %s/%s (fileid %i), xattr not updated: %s", path, name, fileid, str(xattr))

                loginfo = []
                updates = []
                values = []
                xattr_update = False

                if poolname != spool:
                    if _log.getEffectiveLevel() <= logging.INFO:
                        loginfo.append("{0} -> {1}".format(poolname, spool))
                        _log.debug("UPDATE_POOL[%s@%s/%s]: %s -> %s", fileid, path, name, poolname, spool)
                    updates.append('poolname = %s')
                    values.append(spool)
                    if 'pool' in xattr_dict and xattr_dict['pool'] != spool:
                        xattr_dict['pool'] = spool
                        xattr_update = True

                if setname != s_token:
                    if _log.getEffectiveLevel() <= logging.INFO:
                        loginfo.append("{0} -> {1}".format(setname, s_token))
                        _log.debug("UPDATE_TOKEN[%s@%s/%s]: %s -> %s", fileid, path, name, setname, s_token)
                    updates.append('setname = %s')
                    values.append(s_token)
                    if 'accountedspacetokenname' in xattr_dict and xattr_dict['accountedspacetokenname'] != u_token:
                        xattr_dict['accountedspacetokenname'] = u_token
                        xattr_update = True

                if xattr_update:
                    if _log.getEffectiveLevel() <= logging.INFO:
                        loginfo.append("{0} -> {1}".format(json.loads(xattr), xattr_dict))
                        _log.debug("UPDATE_XATTR[%s@%s/%s]: %s -> %s", fileid, path, name, json.loads(xattr), xattr_dict)
                    updates.append('xattr = %s')
                    values.append(json.dumps(xattr_dict))

                _log.info("%supdate replica rowid %i for %s/%s with fileid %i (%s)", dry_run, rowid, path, name, fileid, ', '.join(loginfo))
                if updatedb:
                    values.append(rowid)
                    sql = "UPDATE Cns_file_replica SET {0} WHERE rowid = %s".format(', '.join(updates))
                    ucursor.execute(sql, values)
                    conn.commit()

                updated += 1

        ucursor.close()
        cursor.close()

        _log.info("processed %i records, updated %i", cnt_rows, updated)

    except Exception as e:
        # query in progress that use SSCursor can be killed only by terminating DB connection
        # (closing / deleting cursor lead to retreival of all selected entries from DB)
        del(pathname)
        conn_path.close()
        conn_ss.close()
        conn.close()
        raise

    del(pathname)
    conn_path.close()
    conn_ss.close()
    conn.close()



def fix_dir_size(conn_data, updatelevels, updatedb=False):
    _log.debug("fix_dir_size(updatelevels=%i, updatedb=%s", updatelevels, updatedb)

    dry_run = '' if updatedb else 'dry-run: '

    conn_ss = get_connection(conn_data, 'cns_db', cclass=MySQLdb.cursors.SSCursor)
    conn_path = get_connection(conn_data, 'cns_db')
    pathname = CachedFullPath(conn_path)

    psize = {}
    try:
        # join metadata table with replicas to account size of all file replicas
        sql = 'SELECT metadata.fileid, metadata.parent_fileid, metadata.filesize FROM Cns_file_metadata AS metadata JOIN Cns_file_replica AS replica ON metadata.fileid = replica.fileid'
        cursor = conn_ss.cursor()
        cursor.execute(sql)

        cnt_rows = 0
        while True:
            # retreiving data in chunks dramatically increase performance
            # mostly because of latency associated with each DB query
            rows = cursor.fetchmany(1000)
            if len(rows) == 0: break

            if _log.getEffectiveLevel() < logging.DEBUG:
                _log.debug("fetched %i rows", len(rows))

            fileids = [ row[1] for row in rows ]
            pathnames = pathname.get_ids_multi(fileids)

            for row in rows:
                cnt_rows += 1
                if _log.getEffectiveLevel() < logging.DEBUG:
                    _log.debug("row %i: %s", cnt_rows, str(row))

                fileid, pfileid, filesize = row
                prefix_fileids = pathnames.get(pfileid)
                if prefix_fileids == None:
                    _log.error("skipping fileid %i, unable to reconstruct its parent %i path", fileid, pfileid)
                    continue

                # DPM doesn't update directory size for top two level "/dpm"
                for prefix_fileid in prefix_fileids[2:updatelevels+1]:
                    psize[prefix_fileid] = psize.get(prefix_fileid, 0) + filesize

                if cnt_rows % 1000000 == 0:
                    _log.debug("processed %i records", cnt_rows)

        _log.info("processed %i records", cnt_rows)

        cursor.close()

    except Exception as e:
        # query in progress that use SSCursor can be killed only by terminating DB connection
        # (closing / deleting cursor lead to retreival of all selected entries from DB)
        del(pathname)
        conn_path.close()
        conn_ss.close()
        raise

    conn_ss.close()

    updated = 0
    if len(psize) > 0:
        curr_psize = {}
        conn = get_connection(conn_data, 'cns_db')
        cursor = conn.cursor()
        cursor.execute("SELECT fileid, filesize FROM Cns_file_metadata WHERE fileid IN ({0})".format(','.join([ str(x) for x in psize.keys() ])))
        for row in cursor:
            fileid, filesize = row
            curr_psize[fileid] = filesize
        cursor.close()

        pathnames = pathname.get_multi(psize.keys())

        cursor = conn.cursor()

        for fileid, filesize in sorted(psize.items()):
            if curr_psize.get(fileid) == filesize: continue
            _log.info("%supdate directory %s (id %i) with current size %i to %i", dry_run, pathnames.get(fileid), fileid, curr_psize.get(fileid), filesize)
            if updatedb:
                cursor.execute("UPDATE Cns_file_metadata SET filesize = {0} WHERE fileid = {1}".format(filesize, fileid))
            updated += 1

        conn.commit()
        cursor.close()
        conn.close()

    del(pathname)
    conn_path.close()

    _log.info("updated filesize for %i records", updated)



def fix_spacetokens_size(conn_data, updatedb=False):
    """Update spacetoken usable space according associated directory size."""
    _log.debug("fix_spacetokens_size(updatedb=%s)", updatedb)

    dry_run = '' if updatedb else 'dry-run: '

    conn = get_connection(conn_data, 'cns_db')
    conn_dpm = get_connection(conn_data, 'dpm_db')

    qt = get_quotatoken_data(conn_dpm)

    updated = 0
    path2size = {}
    #for path, poolname, u_token, s_token, t_space, g_space, u_space in qt:
    #    path2size[path] = get_path_size(conn, path)
    for path, poolname, u_token, s_token, t_space, g_space, u_space in qt:
        dirsize = get_path_size(conn, path)
        accounted_dirs = []

        for longer_path, size in sorted(path2size.items(), cmp=lambda x, y: cmp(len(x[0]), len(y[0]))):
            if not longer_path.startswith("%s/" % path): continue
            parent_already_accounted = False
            for accounted_dir in accounted_dirs:
                if longer_path.startswith("%s/" % accounted_dir):
                    parent_already_accounted = True
            if parent_already_accounted: continue
            accounted_dirs.append(longer_path)

        path2size[path] = dirsize
        pathfreespace = t_space - (dirsize - sum([ path2size[x] for x in accounted_dirs ]))

        if u_space != pathfreespace:
            _log.info("%supdate spacetoken %s[%s] free space difference %.02f%%, changing from %i to %i = t_space(%i) - (dirsize(%i) - sum(%s))", dry_run, u_token, path, 100. * (u_space - pathfreespace) / pathfreespace, u_space, pathfreespace, t_space, dirsize, ','.join([ "{0}({1})".format(x, path2size[x]) for x in accounted_dirs ]))
            if updatedb:
                try:
                    cursor = conn_dpm.cursor()
                    cursor.execute("UPDATE dpm_space_reserv SET u_space = %s WHERE path = %s", (pathfreespace, path))
                    conn_dpm.commit()
                finally:
                    cursor.close()
            updated += 1

    conn_dpm.close()
    conn.close()

    _log.info("updated size for %i spacetokens", updated)




#=====================================================================
# main
#=====================================================================
if __name__ == '__main__':
    # basic logging configuration
    streamHandler = logging.StreamHandler(sys.stderr)
    streamHandler.setFormatter(logging.Formatter("%(asctime)s [%(levelname)s](%(module)s:%(lineno)d) %(message)s", "%d %b %H:%M:%S"))
    _log.addHandler(streamHandler)
    _log.setLevel(logging.WARN)

    # parse options from command line
    def opt_set_loglevel(option, opt, value, parser):
        loglevel = option.default
        if value != None:
            loglevel = int({
                'CRITICAL': logging.CRITICAL,
                'DEBUG': logging.DEBUG,
                'ERROR': logging.ERROR,
                'FATAL': logging.FATAL,
                'INFO': logging.INFO,
                'NOTSET': logging.NOTSET,
                'WARN': logging.WARN,
                'WARNING': logging.WARNING,
            }.get(value, value))
        if loglevel < 0:
            # relative log level for multiple -vvv argument
            loglevel += _log.getEffectiveLevel()
            if loglevel < 1: loglevel = 1
        _log.setLevel(loglevel)
        setattr(parser.values, option.dest, loglevel)

    class IndentedHelpFormatterWithEpilogExamples(optparse.IndentedHelpFormatter):
        def format_epilog(self, epilog):
            import textwrap
            if not epilog:
                return ""
            text_width = self.width - self.current_indent
            indent = " "*self.current_indent
            sections = []
            if type(epilog) == str:
                sections.append(textwrap.fill(text, text_width, initial_indent=indent, subsequent_indent=indent))
            else:
                example_section = False
                for epilog_section in epilog:
                    if not epilog_section.startswith('EXAMPLE '):
                        sections.append(textwrap.fill(epilog_section, text_width, initial_indent=indent, subsequent_indent=indent))
                        sections.append('')
                        example_section = False
                    else:
                        if not example_section:
                            sections.append('Examples:')
                            example_section = True
                        sections.append("  {0}{1}".format(indent, epilog_section[len('EXAMPLE '):].replace("\n", "\n{0}".format(indent))))
            return "\n{0}\n".format("\n".join(sections))

    # default config values - log level is not yet set while calling these functions
    guess_nsconfig = guess_config_files()
    guess_updatelevels = get_updatelivels(DEFAULT_DOME_CONFIG)

    # command line arguments
    usage = "usage: %prog [options]"
    epilog = []
    epilog.append(
        "By default this script runs in dry mode without any updates in DPM database. Dry-run "
        "is safe to execute with running DPM services (althought results can be a bit inaccurate "
        "due to updates in database). In case you want to apply updates it is first necessary "
        "to stop all DPM services on your DPM headnode.")
    epilog.append(
        "Description of available DB fixes:"
    )
    epilog.append(
        "* zero-dir -- set to 0 the size of all the directories below first N levels. "
        "Time to apply these updates should be less than a minute and it is reasonable "
        "safe to run this fix even with running DPM services."
    )
    epilog.append(
        "* spacetoken -- update spacetoken information according quotatoken paths "
        "(file uploads with non-SRM protocols and legacy DPM were usually done "
        "without proper association with the spacetoken. If you want to use SRM "
        "protocol with DOME DPM these inconsistencies together with out of sync "
        "spacetoken usage can cause troubles enfocing storage space limits. For "
        "pure DOME DPM without legacy adapter you can skip these updates. This "
        "update examine all records in replica table and it can take several "
        "minutes for DPM with tens of millions files. It should be reasonable "
        "safe to apply this fix even with running DPM services."
    )
    epilog.append(
        "* dir-size -- update directory size for first N levels. To be able to "
        "assign correct directory sizes it is necessary to stop DPM services "
        "before running this fix in \"updatedb\" mode and because to calculate "
        "size of all child objects it is necessary to walk through all records "
        "in file replica table which can take several minutes for DPM with tens "
        "of millions files."
    )
    epilog.append(
        "* spacetoken-size -- synchronize spacetoken used space with quotatoken. "
        "DPM services must be stopped before executing this fix in \"updatedb\" "
        "mode, but counters are updated within a second."
    )
    epilog.append("EXAMPLE # run with default configuration (dry-run without DB updates)")
    epilog.append("EXAMPLE python /usr/bin/dmlite-mysql-dirspaces.py --fix=zero-dir")
    epilog.append("EXAMPLE python /usr/bin/dmlite-mysql-dirspaces.py --fix=spacetoken")
    epilog.append("EXAMPLE python /usr/bin/dmlite-mysql-dirspaces.py --fix=dir-size")
    epilog.append("EXAMPLE python /usr/bin/dmlite-mysql-dirspaces.py --fix=spacetoken-size")
    epilog.append("EXAMPLE ")
    epilog.append("EXAMPLE # minimal updates necessary while moving to DOME DPM")
    epilog.append("EXAMPLE # stop/start all DPM services (including legacy SRM) on CentOS7")
    epilog.append("EXAMPLE systemctl stop httpd rfiod srmv2.2 dpnsdaemon dpm dpm-gsiftp xrootd@dpmredir")
    epilog.append("EXAMPLE python /usr/bin/dmlite-mysql-dirspaces.py --fix=dir-size --updatedb")
    epilog.append("EXAMPLE systemctl start httpd rfiod srmv2.2 dpnsdaemon dpm dpm-gsiftp xrootd@dpmredir")
    epilog.append("EXAMPLE ")
    epilog.append("EXAMPLE # run all consistency checks / updates related to space usage")
    epilog.append("EXAMPLE # recommended for DOME DPM update to fix space usage data")
    epilog.append("EXAMPLE # stop/start all DPM services (including legacy SRM) on CentOS7")
    epilog.append("EXAMPLE systemctl stop httpd rfiod srmv2.2 dpnsdaemon dpm dpm-gsiftp xrootd@dpmredir")
    epilog.append("EXAMPLE python /usr/bin/dmlite-mysql-dirspaces.py --fix=zero-dir,spacetoken,dir-size,spacetoken-size --updatedb")
    epilog.append("EXAMPLE systemctl start httpd rfiod srmv2.2 dpnsdaemon dpm dpm-gsiftp xrootd@dpmredir")
    epilog.append("EXAMPLE # \"zero-dir\" and \"spacetoken\" can be removed")
    epilog.append("EXAMPLE # if dry-run did not show any inconsistencies")
    description = "Update and synchronize space and spacetoken data for DOME DPM."
    parser = optparse.OptionParser(usage=usage, description=description, version="%prog", epilog=epilog, formatter=IndentedHelpFormatterWithEpilogExamples())
    parser.add_option("-v", "--verbose", dest="loglevel", action="callback", callback=opt_set_loglevel, default=-10, help="each \"v\" increases the verbosity level")
    parser.add_option("--debug", dest="loglevel", action="callback", callback=opt_set_loglevel, default=logging.DEBUG, help="set log level to DEBUG")
    #parser.add_option("-v", "--verbose", dest="loglevel", action="callback", callback=opt_set_loglevel, default=logging.DEBUG, help="set log level to DEBUG")
    parser.add_option("-q", "--quiet", dest="loglevel", action="callback", callback=opt_set_loglevel, default=logging.ERROR, help="set log level to ERROR")
    parser.add_option("--log-level", dest="loglevel", action="callback", callback=opt_set_loglevel, type="string", help="set log level, default: %default")
    parser.add_option("--log-file", dest="logfile", metavar="FILE", help="set log file, default: %default")
    parser.add_option("--log-size", dest="logsize", type="int", default=10*1024*1024, help="maximum size of log file, default: %default")
    parser.add_option("--log-backup", dest="logbackup", type="int", default=4, help="number of log backup files, default: %default")
    parser.add_option("-c", "--nsconfig", dest="nsconfig", default=guess_nsconfig, help="NSCONFIG file with sql connection info, default: %default")
    parser.add_option('--dbhost', dest='dbhost', default=None, help="database host, if no NSCONFIG given")
    parser.add_option('--dbuser', dest='dbuser', default=None, help="database user, if no NSCONFIG given")
    parser.add_option('--dbpwd', dest='dbpwd', default=None, help="database password, if no NSCONFIG given")
    parser.add_option('--dbname', dest='dbname', default=DEFAULT_CNS_DB, help="database name, if no NSCONFIG given (default: %default)")
    parser.add_option('--dbdpm', dest='dbdpm', default=DEFAULT_DPM_DB, help="DPM database name, if no NSCONFIG given (default: %default)")
    parser.add_option('--updatedb', dest='updatedb', action='store_true', default=False, help="update original DB with the results of the directory sizes, default: %default")
    parser.add_option('--updatelevels', dest='updatelevels', type='int', default=guess_updatelevels, help="Allows setting the directory size only for the first N levels, default: %default")
    parser.add_option('--fix', dest='fix', default='zero-dir,spacetoken,dir-size,spacetoken-size', help="Apply comma separated DB fixes, default: %default")
    parser.add_option('--skip-spacetokens', dest='skip_spacetokens', action='append', default=[], help="skip fixes for spacetoken, default: %default")


    (options, args) = parser.parse_args()

    if options.logfile == '-':
        _log.removeHandler(streamHandler)
        streamHandler = logging.StreamHandler(sys.stdout)
        streamHandler.setFormatter(logging.Formatter("%(asctime)s [%(levelname)s](%(module)s:%(lineno)d) %(message)s", "%d %b %H:%M:%S"))
        _log.addHandler(streamHandler)
    elif options.logfile != None and options.logfile != '':
        #fileHandler = logging.handlers.TimedRotatingFileHandler(options.logfile, 'midnight', 1, 4)
        fileHandler = logging.handlers.RotatingFileHandler(options.logfile, maxBytes=options.logsize, backupCount=options.logbackup)
        fileHandler.setFormatter(logging.Formatter("%(asctime)s [%(levelname)s](%(module)s:%(lineno)d) %(message)s", "%d %b %H:%M:%S"))
        _log.addHandler(fileHandler)
        _log.removeHandler(streamHandler)

    import socket
    import getpass
    import inspect
    import hashlib
    _log.info("command: %s", " ".join(sys.argv))
    _log.info("script: %s", os.path.abspath(inspect.getfile(inspect.currentframe())))
    _log.info("version: %s", __version__)
    _log.info("sha256: %s", hashlib.sha256(open(__file__).read().encode('utf-8')).hexdigest())
    _log.info("python: %s", str(sys.version_info))
    _log.info("user: %s@%s", getpass.getuser(), socket.gethostname())
    _log.info("system load: %s", str(os.getloadavg()))

    # validate command line options
    fix_names = options.fix.split(',')
    if len(fix_names) == 0:
        _log.error("no fix specified, please --help command line option to get more informations")
        sys.exit(1)
    for fix in fix_names:
        if fix not in [ 'zero-dir', 'spacetoken', 'dir-size', 'spacetoken-size' ]:
            _log.error("unknown fix \"%s\", use --help command line option to get more informations", fix)
            sys.exit(1)

    conn_data = {}
    if options.dbhost == None:
        if options.nsconfig == None:
            _log.error("no database configuration specified, use either nsconfig or db* command line options")
            sys.exit(1)

        conn_data = get_conn_data(options.nsconfig)

    else:
        if options.dbuser == None or options.dbpwd == None:
            _log.error("no database user or password defined as command line options")
            sys.exit(1)

        conn_data['host'] = options.dbhost
        conn_data['user'] = options.dbuser
        conn_data['pass'] = options.dbpwd
        conn_data['cns_db'] = options.dbname
        conn_data['dpm_db'] = options.dbdpm

    _log.info("%s %s for %i levels of the DPM database (host=%s, cns_db=%s, dpm_db=%s)", 'fix' if options.updatedb else 'dry-run', ', '.join(fix_names), options.updatelevels, conn_data['host'], conn_data['cns_db'], conn_data['dpm_db'])

    starttime = datetime.datetime.now()
    if 'zero-dir' in fix_names: fix_zero_dirs(conn_data, options.updatelevels, updatedb=options.updatedb)
    if 'spacetoken' in fix_names: fix_spacetokens_by_path(conn_data, skip=options.skip_spacetokens, updatedb=options.updatedb)
    if 'dir-size' in fix_names: fix_dir_size(conn_data, options.updatelevels, updatedb=options.updatedb)
    if 'spacetoken-size' in fix_names: fix_spacetokens_size(conn_data, updatedb=options.updatedb)
    endtime = datetime.datetime.now()

    _log.info("done (time: %ss)", (endtime-starttime).seconds)
